package geometry.part12;

import geometry.common.Demo;
import geometry.common.DemoFrame;
import geometry.common.DemoPanel;
import geometry.part2.LineSegment;
import geometry.part6.ConvexPolygon;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;

public class VoronoiDemo {
    private static final double SCALE = 520;

    public static void main(String[] args) {
        Demo.launch(VoronoiDemoFrame.class, "ボロノイ図");
    }

    public static class VoronoiDemoFrame extends DemoFrame {
        private static final long serialVersionUID = 1L;

        enum Mode {
            ADD_OR_MOVE, DELETE
        }

        VoronoiPanel panel;
        JToggleButton addOrMoveButton;
        JToggleButton deleteButton;

        List<Point2D> sites;
        Point2D highlightSite;

        @Override
        protected void initialize() {
            sites = new ArrayList<Point2D>();
            for (int i = 0; i < 20; i++) {
                sites.add(new Point2D.Double(Math.random(), Math.random()));
            }

            addOrMoveButton = new JToggleButton("母点の追加 / 移動");
            deleteButton = new JToggleButton("母点の削除");
            ButtonGroup group = new ButtonGroup();
            group.add(addOrMoveButton);
            group.add(deleteButton);
            group.setSelected(addOrMoveButton.getModel(), true);

            JButton deleteAllButton = new JButton("全母点の削除");
            deleteAllButton.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    sites.clear();
                    repaint();
                }
            });

            JToolBar toolBar = new JToolBar();
            toolBar.setFloatable(false);
            toolBar.add(addOrMoveButton);
            toolBar.add(deleteButton);
            toolBar.addSeparator();
            toolBar.add(deleteAllButton);

            panel = new VoronoiPanel();
            panel.addMouseListener(new MouseAdapter() {
                @Override
                public void mousePressed(MouseEvent e) {
                    Point p = e.getPoint();
                    Point2D site = getSiteAt(p);
                    switch (getMode()) {
                    case ADD_OR_MOVE:
                        if (site == null) {
                            Point2D newSite = panel.fromScreen(p);
                            sites.add(newSite);
                            highlightSite = newSite;
                            repaint();
                        }
                        break;
                    case DELETE:
                        if (site != null) {
                            sites.remove(site);
                            repaint();
                        }
                        break;
                    }
                }
            });
            panel.addMouseMotionListener(new MouseMotionAdapter() {
                @Override
                public void mouseMoved(MouseEvent e) {
                    Point2D site = getSiteAt(e.getPoint());
                    if (site != highlightSite) {
                        highlightSite = site;
                        repaint();
                    }
                }

                @Override
                public void mouseDragged(MouseEvent e) {
                    if (getMode() == Mode.ADD_OR_MOVE && highlightSite != null) {
                        Point2D p = panel.fromScreen(e.getPoint());
                        double x = Math.max(0, Math.min(1, p.getX()));
                        double y = Math.max(0, Math.min(1, p.getY()));
                        highlightSite.setLocation(x, y);
                        repaint();
                    }
                }
            });

            Container c = getContentPane();
            c.setLayout(new BorderLayout());
            c.add(toolBar, BorderLayout.NORTH);
            c.add(panel, BorderLayout.CENTER);
        }

        private Point2D getSiteAt(Point p) {
            for (Point2D site : sites) {
                Point2D sp = panel.toScreen(site);
                if (sp.distance(p) <= 8) {
                    return site;
                }
            }
            return null;
        }

        private Mode getMode() {
            if (addOrMoveButton.isSelected()) {
                return Mode.ADD_OR_MOVE;
            } else {
                return Mode.DELETE;
            }
        }

        private class VoronoiPanel extends DemoPanel {
            private static final long serialVersionUID = 1L;

            @Override
            protected void doPaint(Graphics2D g) {
                VoronoiGenerator generator = new VoronoiGenerator();

                List<Point2D> boundary = new ArrayList<Point2D>();
                boundary.add(new Point2D.Double(0, 0));
                boundary.add(new Point2D.Double(1, 0));
                boundary.add(new Point2D.Double(1, 1));
                boundary.add(new Point2D.Double(0, 1));

                List<ConvexPolygon> result = generator.execute(
                        new ConvexPolygon(boundary), sites);

                g.setPaint(Color.BLACK);
                for (ConvexPolygon p : result) {
                    drawPolygon(g, p);
                }
                for (Point2D site : sites) {
                    Point2D p = toScreen(site);
                    double r = site == highlightSite ? 8 : 4;
                    g.fill(new Ellipse2D.Double(p.getX() - r, p.getY() - r,
                            2 * r, 2 * r));
                }
            }

            void drawPolygon(Graphics2D g, ConvexPolygon polygon) {
                if (polygon == null) {
                    return;
                }
                for (int i = 0; i < polygon.getEdgeCount(); i++) {
                    LineSegment segment = polygon.getEdge(i);
                    Point2D p1 = toScreen(segment.x1, segment.y1);
                    Point2D p2 = toScreen(segment.x2, segment.y2);
                    g.draw(new Line2D.Double(p1, p2));
                }
            }

            Point2D toScreen(Point2D p) {
                return toScreen(p.getX(), p.getY());
            }

            Point2D toScreen(double x, double y) {
                double cx = getWidth() / 2;
                double cy = getHeight() / 2;
                return new Point2D.Double(cx + (x - 0.5) * SCALE, cy
                        + (y - 0.5) * SCALE);
            }

            Point2D fromScreen(Point p) {
                double cx = getWidth() / 2;
                double cy = getHeight() / 2;
                return new Point2D.Double((p.x - cx) / SCALE + 0.5, (p.y - cy)
                        / SCALE + 0.5);
            }
        }
    }
}
